/*
 * nzb
 *
 * Copyright (C) 2004-2006 Mattias Nordstrom <matta at ftlight net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * Authors:
 *   Mattias Nordstrom <matta at ftlight net>
 *
 * $Id: mainwindow.cpp,v 1.18 2005/10/16 12:03:32 mnordstr Exp $
 *   This file provides the GUI.
 */

#include "mainwindow.h"
#include "downloader.h"
#include "optionsdlg.h"
#include "decoder.h"
#include "output.h"

MainWindow::MainWindow( QWidget* parent ) :
        QMainWindow( parent ),
        Ui::MainWindow(),
        nzblist(),
        downloaders(),
        decoder( NULL ),
        output( NULL ),
        file_lock(),
        //dc_done(),
        totalSpeed( NULL ),
        totalSize( NULL ),
        totalProgress( NULL ),
        shutting_down( false ),
        download_complete( false ),
        config( NULL )
{
        this->Ui::MainWindow::setupUi(this);
        this->list->addAction( this->action_EnableDownload );
        this->list->addAction( this->action_DisableDownload );
        // connect signals
        connect(this->action_Open, SIGNAL(triggered()), this, SLOT(open()));
        connect(this->action_Exit, SIGNAL(triggered()), this, SLOT(close()));
        connect(this->action_Sort, SIGNAL(triggered()), this, SLOT(sortList()));
        connect(this->action_Clear, SIGNAL(triggered()), this, SLOT(clearList()));
        connect(this->action_About, SIGNAL(triggered()), this, SLOT(about()));
        connect(this->action_Options, SIGNAL(triggered()), this, SLOT(options()));
        connect(this->action_Start, SIGNAL(triggered()), this, SLOT(start()));
        connect(this->action_Stop, SIGNAL(triggered()), this, SLOT(stop()));
        connect(this->action_Stream, SIGNAL(triggered()), this, SLOT(stream()));
        connect(this->action_Pause, SIGNAL(triggered()), this, SLOT(pause()));
        connect(this->action_Resume, SIGNAL(triggered()), this, SLOT(resume()));
        connect(this->action_EnableDownload, SIGNAL(triggered()), this, SLOT(enableDownloadOfItems()));
        connect(this->action_DisableDownload, SIGNAL(triggered()), this, SLOT(disableDownloadOfItems()));

        // statusbar
        this->statusBar()->showMessage(tr("Ready"));

        totalSpeed = new QLabel("0 kB/s");
        totalSize = new QLabel("0 MB");
        totalProgress = new QProgressBar();
        totalProgress->setRange(0, 1);
        totalProgress->setValue(0);

        this->statusBar()->addPermanentWidget(totalSpeed);
        this->statusBar()->addPermanentWidget(totalProgress);
        this->statusBar()->addPermanentWidget(totalSize);

        // read settings
        this->setConfig(this->loadSettings());

        // configure
        const int connections = config->value("server/connections").toInt();
        this->processes->setRowCount(connections + 1);

        // Downloaders
        for (int i = 0; i < connections; ++i) {
                downloaders.append( new Downloader(&nzblist, i, &file_lock, this->list, this) );
                connect(this, SIGNAL(fullStop()), downloaders[i], SLOT(stop()));
                connect(this, SIGNAL(requestPause()), downloaders[i], SLOT(pause()));
                connect(this, SIGNAL(requestResume()), downloaders[i], SLOT(resume()));
                this->processes->setItem(i, 0, new QTableWidgetItem(QIcon(":/images/connect_no.png"), "Downloader #"+QString::number(i+1)));
                this->processes->setItem(i, 2, new QTableWidgetItem("0 kB/s"));
                this->processes->item(i, 2)->setTextAlignment(Qt::AlignRight);
        }

        // Decoder
        {
                decoder = new Decoder(&nzblist, connections, &file_lock, this);
                connect(this, SIGNAL(processDecoder()), decoder, SLOT(processFiles()));
                connect(this, SIGNAL(fullStop()), decoder, SLOT(stop()));
                this->processes->setItem(connections, 0, new QTableWidgetItem(QIcon(":/images/decoder.png"), "Decoder #1"));
        }

        // Output
        {
                output = new Output(&nzblist, connections+1, &file_lock, this);
                connect(this, SIGNAL(processOutput()), output, SLOT(process()));
                connect(this, SIGNAL(fullStop()), output, SLOT(stop()));
        }

        const int drows = this->processes->rowCount();
        for( int row = 0; row < drows; ++row ) {
                this->processes->setRowHeight(row, 16);
        }
        //this->processes->resizeRowsToContents();
        this->Tabs->setCurrentIndex(0);
}

MainWindow::~MainWindow() {
         for( int i = 0; i < downloaders.size(); ++i ) {
                 delete downloaders[i];
         }
         delete decoder;
         delete output;
         delete config;
}

void MainWindow::loadFile(const QString &fileName)
{
        QFile f(fileName);
        nzblist.importNzb(&f);

        if (config->value("list/autosort").toBool()) {
                this->sortList();
        } else {
                this->populateList();
        }

        if (nzblist.getList()->count()) {
                this->action_Start->setEnabled(true);
                this->action_Stream->setEnabled(true);
        }

        this->statusBar()->showMessage( tr("Loaded nzb document %1").arg(fileName), 2000 );
}

void MainWindow::setConfig(QSettings *config)
{
        delete this->config;
        this->config = config;

        QPoint pos = config->value("mainwindow/pos", QPoint(200, 200)).toPoint();
        QSize size = config->value("mainwindow/size", QSize(400, 400)).toSize();
        this->resize(size);
        this->move(pos);
        this->split->restoreState(config->value("mainwindow/split").toByteArray());

        this->list->horizontalHeader()->resizeSection(0, config->value("mainwindow/listcol0width", 50).toInt());
        this->list->horizontalHeader()->resizeSection(1, config->value("mainwindow/listcol1width", 50).toInt());
        this->list->horizontalHeader()->resizeSection(2, config->value("mainwindow/listcol2width", 50).toInt());
        this->list->horizontalHeader()->resizeSection(3, config->value("mainwindow/listcol3width", 50).toInt());
        this->processes->horizontalHeader()->resizeSection(0, config->value("mainwindow/dlistcol0width", 50).toInt());
        this->processes->horizontalHeader()->resizeSection(1, config->value("mainwindow/dlistcol1width", 50).toInt());
        this->processes->horizontalHeader()->resizeSection(2, config->value("mainwindow/dlistcol2width", 50).toInt());

        // Fix for situation when the default value isn't honored.
        if (this->list->columnWidth(0) == 0) this->list->horizontalHeader()->resizeSection(0, 50);
        if (this->list->columnWidth(1) == 0) this->list->horizontalHeader()->resizeSection(1, 50);
        if (this->list->columnWidth(2) == 0) this->list->horizontalHeader()->resizeSection(2, 50);
        if (this->list->columnWidth(3) == 0) this->list->horizontalHeader()->resizeSection(3, 50);
        if (this->processes->columnWidth(0) == 0) this->processes->horizontalHeader()->resizeSection(0, 50);
        if (this->processes->columnWidth(1) == 0) this->processes->horizontalHeader()->resizeSection(1, 50);
        if (this->processes->columnWidth(2) == 0) this->processes->horizontalHeader()->resizeSection(2, 50);
}

void MainWindow::downloadEvent(QString message, int row, int type)
{
        if (type == 0) {
                this->processes->model()->setData(this->processes->model()->index(row, 1), message);
        } else if (type == 2) {
                this->list->model()->setData(this->list->model()->index(row, 1), (this->list->item(row, 1)->text()).toInt()+1);
                totalProgress->setValue(totalProgress->value() + 1);
                setWindowTitle(tr("nzb") + " - " + QString::number(((double)totalProgress->value() / (double)totalProgress->maximum() * 100.00), 'f' , 0) + "%");

                if (this->list->item(row, 1)->text() == "1") {
                        if (config->value("list/hidecompleted").toBool()) {
                                this->list->scrollToItem(this->list->item(row, 1));
                        }
                }

                emit processDecoder();

        } else if (type == 3) {
                this->processes->model()->setData(this->processes->model()->index(row, 1), message);
        } else if (type == 4) {
                this->processes->model()->setData(this->processes->model()->index(row, 2), message+" kB/s");

                int total_speed = 0;
                for (int i=0; i<downloaders.size(); i++) {
                        QString cur_speed = this->processes->item(i, 2)->text();
                        total_speed += cur_speed.mid(0, cur_speed.length() - 5).toInt();
                }
                totalSpeed->setText(QString::number(total_speed)+" kB/s");
        } else if (type == 5) {
                this->processes->model()->setData(this->processes->model()->index(row, 1), message);
                QTimer::singleShot(60000, this, SLOT(restartDownloaders()));
        } else if (type == 6) {
                this->processes->model()->setData(this->processes->model()->index(row, 1), message);
        }
}

void MainWindow::decodeEvent(QString message, int row, int type)
{
        this->processes->model()->setData(this->processes->model()->index(row, 1), message);

        if (type == 1) {
                emit processOutput();
        } else if (type == 2) {
                QTimer::singleShot(1000, decoder, SLOT(processFiles()));
        } else if (type == 3) {
                QTimer::singleShot(0, output, SLOT(process()));
        }
}

void MainWindow::outputEvent(QString message, int type)
{
        if (type == 1) {
                download_complete = true;
                QTimer::singleShot(200, this, SLOT(stop()));
        } else if (type == 2) {
                download_complete = true;
                QTimer::singleShot(200, this, SLOT(stop()));
                QMessageBox *msgbox = new QMessageBox("nzb Error", message, QMessageBox::Critical, QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
                msgbox->show();
        }
}

void MainWindow::closeEvent(QCloseEvent *event)
{
        shutting_down = true;
        if (this->action_Stop->isEnabled()) {
                this->stop();
                if (!shutting_down) {
                        event->ignore();
                        return;
                }
        }
        this->saveSettings();
        event->accept();
}

void MainWindow::open()
{
        QString defaultPath = config->value("mainwindow/openpath").toString();
        QString fileName = QFileDialog::getOpenFileName(this, "Open NZB File", defaultPath, "nzb Files (*.nzb)\nAll Files (*)");
        if (!fileName.isEmpty()) {
            QFileInfo fi(fileName);
            config->setValue("mainwindow/openpath", fi.absolutePath());
                loadFile(fileName);
        }
}

void MainWindow::about()
{
        QMessageBox::about(this, tr("About nzb"), tr("nzb v%1\n\nnzb is a Usenet binary downloader.\n\nSee http://www.nzb.fi/").arg(NZB_VERSION));
}

void MainWindow::start()
{
        download_complete = false;

        action_Start->setEnabled(false);
        action_Stop->setEnabled(true);
        action_Stream->setEnabled(false);
        action_Pause->setEnabled(true);
        action_Resume->setEnabled(false);

        for( int i = 0; i < downloaders.size(); ++i ) {
                downloaders[i]->start(QThread::LowPriority);
        }

        decoder->start();

        output->setStream(false);
        output->start(QThread::TimeCriticalPriority);
}

void MainWindow::stop()
{
        if (!download_complete) {
                QMessageBox msgBox;
                msgBox.setText("If you stop the current file list will be cleared. You can use the pause action to temporary stop the download.");
                msgBox.setInformativeText("Are you sure you want to stop and clear the list?");
                msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
                msgBox.setDefaultButton(QMessageBox::No);
                msgBox.setIcon(QMessageBox::Question);
                int ret = msgBox.exec();
                if (ret == QMessageBox::No) {
                        shutting_down = false;
                        return;
                }
        }

        action_Start->setEnabled(true);
        action_Stop->setEnabled(false);
        action_Stream->setEnabled(true);
        action_Pause->setEnabled(false);
        action_Resume->setEnabled(false);

        qDebug("Stopping");

        // tell workers to stop
        emit fullStop();
        // stop threads
        for( int i = 0; i < downloaders.size(); ++i) {
                downloaders[i]->wait();
        }

        decoder->wait();

        //output->wait();

        this->clearData();

        this->setWindowTitle(tr("nzb"));
}

void MainWindow::pause()
{
        action_Start->setEnabled(false);
        action_Stop->setEnabled(true);
        action_Stream->setEnabled(false);
        action_Pause->setEnabled(false);
        action_Resume->setEnabled(true);
        emit requestPause();
}

void MainWindow::resume()
{
        action_Start->setEnabled(false);
        action_Stop->setEnabled(true);
        action_Stream->setEnabled(false);
        action_Pause->setEnabled(true);
        action_Resume->setEnabled(false);
        emit restartDownloaders();
}

void MainWindow::stream()
{
        if (getConfig()->value("output/mediaplayer").toString().isEmpty()) {
                QMessageBox msgBox;
                msgBox.setText("No media player has been specified in output settings.");
                msgBox.setInformativeText("Please specify an application for streaming before using this function.");
                msgBox.setStandardButtons(QMessageBox::Ok);
                msgBox.setDefaultButton(QMessageBox::Ok);
                msgBox.setIcon(QMessageBox::Critical);
                msgBox.exec();

                return;
        }

        action_Start->setEnabled(false);
        action_Stop->setEnabled(true);
        action_Stream->setEnabled(false);
        action_Pause->setEnabled(true);
        action_Resume->setEnabled(false);

        for( int i = 0; i < downloaders.size(); ++i ) {
                downloaders[i]->start(QThread::LowPriority);
        }

        decoder->start();

        output->setStream(true);
        output->start(QThread::TimeCriticalPriority);
}

void MainWindow::options()
{
        OptionsDlg dl(this);
        dl.exec();
}

void MainWindow::sortList()
{
        this->list->clearContents();
        this->list->setRowCount(0);

        nzblist.sortList();

        this->populateList();
}

void MainWindow::restartDownloaders()
{
        for( int i = 0; i < downloaders.size(); ++i ) {
                if (downloaders[i]->isFinished()) {
                    downloaders[i]->start(QThread::LowPriority);
                }
        }
}

void MainWindow::disableDownloadOfItems()
{
        // get list of selected items
        const QList<QTableWidgetSelectionRange> items = this->list->selectedRanges();
        // disable items
        for( QList<QTableWidgetSelectionRange>::const_iterator it = items.begin(); it != items.end(); ++it ) {
                const QTableWidgetSelectionRange range = *it;
                for( int i = range.topRow(); i <= range.bottomRow(); ++i ) {
                        this->list->item(i, 0)->setCheckState(Qt::Unchecked);
                }
        }
}

void MainWindow::enableDownloadOfItems()
{
        // get list of selected items
        const QList<QTableWidgetSelectionRange> items = this->list->selectedRanges();
        // disable items
        for( QList<QTableWidgetSelectionRange>::const_iterator it = items.begin(); it != items.end(); ++it ) {
                const QTableWidgetSelectionRange range = *it;
                for( int i = range.topRow(); i <= range.bottomRow(); ++i ) {
                        this->list->item(i, 0)->setCheckState(Qt::Checked);
                }
        }
}

QSettings* MainWindow::loadSettings()
{
        QSettings* c = NULL;
        if (QFile::exists("nzb.ini")) {
                c = new QSettings("nzb.ini", QSettings::IniFormat);
        } else {
                c = new QSettings("nzb", "nzb");
        }
        return c;
}

void MainWindow::saveSettings()
{
        // save settings of mainwindow
        config->setValue("mainwindow/pos", this->pos());
        config->setValue("mainwindow/size", this->size());
        config->setValue("mainwindow/split", this->split->saveState());
        config->setValue("mainwindow/listcol0width", this->list->columnWidth(0));
        config->setValue("mainwindow/listcol1width", this->list->columnWidth(1));
        config->setValue("mainwindow/listcol2width", this->list->columnWidth(2));
        config->setValue("mainwindow/listcol3width", this->list->columnWidth(3));
        config->setValue("mainwindow/dlistcol0width", this->processes->columnWidth(0));
        config->setValue("mainwindow/dlistcol1width", this->processes->columnWidth(1));
        config->setValue("mainwindow/dlistcol2width", this->processes->columnWidth(2));
        config->sync();
}

void MainWindow::clearData()
{
        this->list->clearContents();
        this->list->setRowCount(0);
//         this->list->setHorizontalHeaderLabels((QStringList)"Subject" << "Progress" << "Parts" << "Size");

        nzblist.clearList();
        file_lock.unlock();

        totalSize->setText("0 MB");
        totalProgress->setRange(0,1);
        totalProgress->setValue(0);

        action_Start->setEnabled(false);
        action_Stop->setEnabled(false);
        action_Stream->setEnabled(false);
        action_Pause->setEnabled(false);
        action_Resume->setEnabled(false);
}

void MainWindow::populateList()
{
        // row counts
        const int old_rows = this->list->rowCount();
        const int new_rows = nzblist.getList()->count();
        this->list->setRowCount(new_rows);

        for( int i = old_rows; i < new_rows; ++i ) {
                QTableWidgetItem *newItem = new QTableWidgetItem(nzblist.getFile(i)->getSubject());
                QTableWidgetItem *newItem_prog = new QTableWidgetItem("0");
                QTableWidgetItem *newItem_tot = new QTableWidgetItem(QString::number(nzblist.getFile(i)->getSegments()->size()));

                QTableWidgetItem *newItem_size;
                int file_bytes = nzblist.getFile(i)->getBytes();

                /*if (file_bytes >= 1e09) {
                        newItem_size = new QTableWidgetItem(QString::number((file_bytes / 1073741824))+" GB");
                } else*/
                if (file_bytes >= 1e06) {
                        newItem_size = new QTableWidgetItem(QString::number((file_bytes / 1048576))+" MB");
                } else if (file_bytes >= 1e03) {
                        newItem_size = new QTableWidgetItem(QString::number((file_bytes / 1024))+" kB");
                } else {
                        newItem_size = new QTableWidgetItem(QString::number(file_bytes)+" bytes");
                }

                newItem_prog->setTextAlignment(Qt::AlignRight);
                newItem_tot->setTextAlignment(Qt::AlignRight);
                newItem_size->setTextAlignment(Qt::AlignRight);
                newItem->setCheckState(Qt::Checked);

                this->list->setItem(i, 0, newItem);
                this->list->setItem(i, 1, newItem_prog);
                this->list->setItem(i, 2, newItem_tot);
                this->list->setItem(i, 3, newItem_size);

                if (config->value("list/par_uncheck").toBool()) {
                        QRegExp rx("*vol*.par2");
                        rx.setPatternSyntax(QRegExp::Wildcard);
                        rx.setCaseSensitivity(Qt::CaseInsensitive);
                        if (nzblist.getFile(i)->getSubject().contains(rx)) {
                                newItem->setCheckState(Qt::Unchecked);
                        }
                }

                //this->list->resizeRowToContents(i);
                this->list->setRowHeight(i, 16);
        }

        qint64 total_size = nzblist.totalSize();

        if (total_size >= 1e09) {
                totalSize->setText(QString::number((double)((double)(total_size / 1048576) / 1000), 'f', 2)+" GB");
        } else if (total_size >= 1e06) {
                totalSize->setText(QString::number((total_size / 1048576))+" MB");
        } else if (total_size >= 1e03) {
                totalSize->setText(QString::number((total_size / 1024))+" kB");
        } else {
                totalSize->setText(QString::number(total_size)+" bytes");
        }

        totalProgress->setRange(0, nzblist.totalParts());
}
